/*
 *  Linux driver for the Catweasel MK4 by Individual Computers
 *
 *  Copyright (C) 2000-2004  Jarno Paananen <jpaana@s2.org>
 *  Copyright (C) 2001-2005  Simon White    <s_a_white@email.com>
 *  COpyright (C) 2005-2006  Antti Lankila  <alankila@bel.fi>
 *
 *  Original HardSID driver was based on the Linux kernel RTC driver.
 *  This driver is based on the experimental CWMK4 branch from hardsid sourceforge
 *  repository. Project has been forked 29.12.2005.
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 */

#define HSID_VERSION    "0.17-dev-cwmk4-kernel3.x-jvaltane"
#define HSID_NAME       "hardsid"

#include <linux/version.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/reboot.h>
#if 1
#include <linux/sched.h>
#endif
#include <asm/io.h>

#ifdef CONFIG_FW_LOADER_MODULE
#   define CONFIG_FW_LOADER CONFIG_FW_LOADER_MODULE
#endif

#include <linux/firmware.h>

#include "hardsid.h"

#define HSID_MAX_CARDS 1
#define HSID_MAX_SIDS_PER_CARD 2
#define HSID_MAX_SIDS (HSID_MAX_CARDS * HSID_MAX_SIDS_PER_CARD)

/* Nothing configurable found below */
static wait_queue_head_t q_wait;
static volatile unsigned long delay_drift = 0;

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

#define PRIVATEDATA(file) ((file)->private_data)

#define CW4_BUFFER_SIZE 1024 /* size of hardware buffer for the ioctl, but problematic */

#if 1
# ifndef pci_module_init
#  define pci_module_init pci_register_driver
# endif
# ifndef SA_SHIRQ
#  define SA_SHIRQ IRQF_SHARED
# endif
#endif

/* These IDs are not registered and may belong to others */
#define PCI_VENDOR_ID_INDIVIDUAL 0xe159
#define PCI_DEVICE_ID_INDIVIDUAL_CWMK 0x0001
#define PCI_SUBSYSTEM_VENDOR_ID_INDIVIDUAL_CWMK4 0x5213
#define PCI_SUBSYSTEM_DEVICE_ID_INDIVIDUAL_CWMK4 0x0002
                                        
static __initdata struct pci_device_id id_table[] = {
    { PCI_VENDOR_ID_INDIVIDUAL, PCI_DEVICE_ID_INDIVIDUAL_CWMK,
      PCI_SUBSYSTEM_VENDOR_ID_INDIVIDUAL_CWMK4,
      PCI_SUBSYSTEM_DEVICE_ID_INDIVIDUAL_CWMK4, 0, 0, 0 },
    { 0, }
};

static const char * const sid_card[] =
{
  "",
  "HardSID",
  "HardSID Quattro",
  "Catweasel MK3",
  "HardSID PCI",
  "HardSID PCI Quattro",
  "Catweasel MK4"
};

typedef struct sid_d
{
    unsigned short       port;
    unsigned char        chip;
    sid_type             type;
    sid_card_type        card;
    struct pci_dev*      pcidev;
    int                  hwbuff;

/* ReSID-style faked reads */
    unsigned char        fakedRead;

/* SID state */
    unsigned char        sidState[32];
/* accumulator state */
    unsigned int         accumulator;
/* filter control disabled? */
    unsigned int         nofilter;
    
/* read write callbacks */
    void                (*poke) (struct sid_d *sid, unsigned char reg,
                                 unsigned char value);
    unsigned char       (*peek) (struct sid_d *sid, unsigned char reg);
} sid_d;

/* Function prototypes */
static loff_t  hsid_llseek     (struct file *file, loff_t offset,
                                int origin);
static ssize_t hsid_read       (struct file *file, char *buf,
                                size_t count, loff_t *ppos);
static int     hsid_ioctl      (struct inode *inode, struct file *file,
                                unsigned int cmd, unsigned long arg);
static void    hsid_delay      (sid_d *sid, unsigned int delay); /* non-immediate delay */
static void    hsid_write_pr   (sid_d *sid, __u32 cmd);
static void    hsid_delay_pr   (sid_d *sid, __u32 delay); /* immediate delay */
static int     hsid_read_proc  (char *page, char **start, off_t off,
                                int count, int *eof, void *data);
static int     hsid_release_pr (struct file *file);
static void    clock_sid_emulation(sid_d *sid, int clocks);

/* Global data */
static spinlock_t hsid_open_lock;
static spinlock_t hsid_bank_lock;
static spinlock_t hsid_delay_update;
static struct fasync_struct *hsid_async_queue;
sid_d *sid_data[HSID_MAX_SIDS];
static int sid_numSIDs;
static int sid_open;
static sid_d sid_setup;
static int pci_allocated;

/* End of Globals */

/* push value into the SID chip */
static void m_fifo_in (struct sid_d *sid, unsigned char cmd)
{
    unsigned long flags;
    DEFINE_WAIT(fwait);
    
    if (! sid->hwbuff) {
        sid->hwbuff = (0xff ^ inb(sid->port + 0xc4)) << 2;
        if (! sid->hwbuff) {
            prepare_to_wait(&q_wait, &fwait, TASK_INTERRUPTIBLE);
            /* The value here should be about 0xff but maybe we were delayed
             * by something. We get the IRQ when 0x80 switches to 0x7f. */
            if (inb(sid->port + 0xc4) >= 0x80)
                schedule();
            finish_wait(&q_wait, &fwait);
            sid->hwbuff = (0xff ^ inb(sid->port + 0xc4)) << 2;
        }
    }

    sid->hwbuff -= 1;
    spin_lock_irqsave(&hsid_bank_lock, flags);
    outb(cmd, sid->port + 0xc0);
    spin_unlock_irqrestore(&hsid_bank_lock, flags);
}

/* these poke commands don't work after FIFO operation has been initiated */
static void hsid_poke_cwmk3 (sid_d *sid, unsigned char reg,
                             unsigned char value)
{
    unsigned short port = sid->port;
    outb(value, port + 0xd8);
    outb(reg | (sid->chip << 7), port + 0xdc);
    sid->sidState[reg] = value;
}

static unsigned char hsid_peek_cwmk3 (sid_d *sid, unsigned char reg)
{
    unsigned short port = sid->port;
    outb(reg | 0x20 | (sid->chip << 7), port + 0xdc);
    udelay(2);
    return inb(port + 0xd8);
}

/* the catweasel delay is always one longer than the value written. The 6th bit selects
 * 64-cycle multiplier for the delay value, so 0x7f waits for 0x40 * 64 cycles. */
static void hsid_delay_pr(sid_d *sid, __u32 delay)
{
    /* issue 4096 cycle delays until the remaining delay is less than that */
    while (delay >= 0x40 * 64) {
        m_fifo_in(sid, 0x7f);
        delay -= 0x40 * 64;
    }
    /* issue the 64 * N delay */
    if (delay >= 0x40) {
        m_fifo_in(sid, 0x40 | ((delay >> 6) - 1));
        delay &= 0x3f;
    }
    /* issue any remaining delay */
    if (delay)
        m_fifo_in(sid, delay-1);

    /* the value 0x3f is reserved and may not be emitted */
    return;
}

/* poke-based reset */
static void hsid_reset_poke(sid_d *sid, unsigned char vol)
{
    int i;
    
    for (i = 0; i < 0x18; i++ ) {
        sid->poke(sid, i, 0);
        udelay(4);
    }
    /* Set the volume */
    sid->poke(sid, i, vol);
    udelay(4);

    return;
}

static void hsid_reset_stream(sid_d *sid, unsigned char vol)
{
    int i;
    for (i = 0; i < 0x18; i++ ) {
        hsid_write_pr(sid, (12 << 16) | (i << 8) | 0);
    }
    /* Set the volume */
    hsid_write_pr(sid, (12 << 16) | (i << 8) | vol);

    return;
}

/*
 *    Now all the various file operations that we export.
 */

static loff_t hsid_llseek(struct file *file, loff_t offset, int origin)
{
    switch(origin)
    {
    case 0:
        file->f_pos = offset;
        return file->f_pos;
    case 1:
        file->f_pos += offset;
        return file->f_pos;
    default:
        return -EINVAL;
    }
}

void hsid_delay(sid_d *sid, unsigned int delay)
{
    /* for pure delays, we merge them together to issue them at a later time */
    spin_lock(&hsid_delay_update);
    delay_drift -= delay;
    spin_unlock(&hsid_delay_update);
    clock_sid_emulation(sid, delay);
}

static void hsid_write_pr(sid_d *sid, __u32 cmd)
{
    unsigned char  reg  = (unsigned char) ((cmd >> 8) & 0x1f);
    int delay = cmd >> 16;
    clock_sid_emulation(sid, delay); /* delay before write */

    /* volume register writes get altered when nofilter is enabled */
    if (sid->nofilter && reg == 0x17)
        cmd &= 0xf8;
    sid->sidState[reg] = cmd;
        
    /* The plan here is to wait in the minimum 4-5 cycles between commands.
     * We do this by ensuring that the delay value is at least 10, which breaks
     * up as follows:
     *
     * leading delay that is minimum of 4 cycles long
     * first sid write (1 cycle)
     * 4 cycle delay
     * second sid write (1 cycle)
     *
     * In order to not slip in our timings (by too much) we remember whether we are
     * lagging behind in the delay_drift variable and compensate in the future when
     * longer delays occur.
     *
     * I've noticed that the Catweasel doesn't always manage to do the write
     * properly in just 4 cycles, sometimes 5 cycles are required or the write
     * is simply lost. However, we can't afford 5 cycles because the SID is quite
     * cycle-sensitive and songs begin to play wrong if the minimum delay is
     * 12 cycles. We get by 10 cycles with the skin of our teeth.
     *
     * It is possible these problems are due to having to set the chip select pin
     * which probably requires some kind of stabilizing delay.
     */

    spin_lock(&hsid_delay_update); 
    delay -= delay_drift;
    delay_drift = 0;
    spin_unlock(&hsid_delay_update);
    
    if (0 && sid_numSIDs == 2)  {
        /* this implements stream write to both chips (with same stream) */
    
        /* We try to use 5 clock delays when possible, but if the stream
         * pushes us around, we give up and use only 4 cycles. */
        int cmd_delay = 5;
        /* we automatically issue the other delay as 5-clock delay and the other
         * as 4-clock delay if the delay value is 11 */
        if (delay <= 10) {
            cmd_delay = 4;
            spin_lock(&hsid_delay_update); 
            delay_drift += 10 - delay;
            spin_unlock(&hsid_delay_update); 
            delay = 10;
        }

        /* wait for the delay minus the time we will wait other ways */
        hsid_delay_pr(sid, delay - (1 + cmd_delay + 1));
        /* 1 cycle */
        m_fifo_in(sid, 0x80 | (0 << 5) | reg);
        m_fifo_in(sid, cmd & 0xff);
        /* cmd_delay cycles */
        hsid_delay_pr(sid, cmd_delay);
        /* 1 cycle */
        m_fifo_in(sid, 0x80 | (1 << 5) | reg);
        m_fifo_in(sid, cmd & 0xff);

    } else {
        /* single chip write */

        if (delay < 1) {
            spin_lock(&hsid_delay_update); 
            delay_drift += 1 - delay;
            spin_unlock(&hsid_delay_update);
            delay = 1;
        }
        hsid_delay_pr(sid, delay - 1);
        m_fifo_in(sid, 0x80 | ((cmd >> 8) & 0x20) | reg);
        m_fifo_in(sid, cmd & 0xff);
    }
    return;
}

static ssize_t hsid_write(struct file * file, const char * buffer,
                          size_t count, loff_t *ppos)
{
    int             ret = 0;
    size_t          bytes;
    __u32           buf;
    const char      *p = buffer;
    size_t          c = count;
    sid_d*          sid = PRIVATEDATA(file);

    while (c > 0) {
        bytes = MIN(c, sizeof(buf));

        bytes -= copy_from_user(&buf, p, bytes);
        if (bytes == 0) {
            printk("<1>" HSID_NAME ": warning: short write (0 B) to /dev/sid.\n");
            ret = -EFAULT;
            break;
        }
        c -= bytes;
        p += bytes;

        /* We want multiples of 4 bytes here */
        if (bytes != 4) {
            printk("<1>" HSID_NAME ": warning: short write (%d B) to /dev/sid.\n", (int) bytes);
            break;
        }
        
        /* Support ReSID-based faked reads */
        sid->fakedRead  = (unsigned char) (buf & 0xff);

        /* Command structure:
           bits 31-16: 16 bit delay timer value in C64 clock cycles
           bits 15-13: reserved, keep zero
           bits 12-8:  SID register number
           bits 7-0:   Data
        */
        hsid_write_pr(sid, buf);
    }

    if (p == buffer) {
        return ret;
    } else {
        file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
        mark_inode_dirty(file->f_dentry->d_inode);
        return p - buffer;
    }
}

/* This is the lousiest OSC3 emulation ever, but it is enough for Polly Tracker. */
static void clock_sid_emulation(sid_d *sid, int clocks)
{
    int freq = (sid->sidState[7 * 2 + 1] << 8) | sid->sidState[7 * 2 + 0];
    sid->accumulator += freq * clocks;
    sid->accumulator &= 0xffffff;
}

static unsigned char hsid_read_pr(sid_d *sid, __u32 cmd)
{
    int reg = (cmd >> 8) & 0xff;
    hsid_delay(sid, cmd >> 16);
    
    if (reg == 0x19) {
        printk("<1>" HSID_NAME ": warning: reading from POTX not supported.\n");
        return 0;
    }
    if (reg == 0x1A) {
        printk("<1>" HSID_NAME ": warning: reading from POTY not supported.\n");
        return 0;
    }

    /* We can't go to the chip and actually do the read because the FIFO is in
     * the way. It's not only preventing us from accessing the chip, but also introduces
     * a latency factor. The userland application is usually some 1 frame further
     * ahead than we are, and we can't read the correct value without draining the
     * FIFO first.
     */
    if (reg == 0x1B) {
        /* for Polly Tracker */
        if ((sid->sidState[7 * 2 + 4] & 0xf0) == 0x40) /* pulse wave */
            return sid->accumulator & 0x800000 ? 128 : 127;
        
        printk("<1>" HSID_NAME ": warning: player read OSC3/RND (state=%02x). Player might not work.\n", sid->sidState[7 * 2 + 4]);
        return 0;
    }
    if (reg == 0x1C) {
        printk("<1>" HSID_NAME ": warning: player read ENV3. Player might not work.\n");
        return 0;
    }
    
    return sid->fakedRead;
}

static ssize_t hsid_read(struct file *file, char *buf,
                         size_t count, loff_t *ppos)
{
    unsigned long i   = *ppos;
    sid_d*        sid = PRIVATEDATA(file);

    /* We want only 1 byte reads */
    if (count != 1)
        return -EFAULT;

    if (i > 0x1f)
        return -EFAULT;

    if (! access_ok(VERIFY_WRITE, buf, count))
        return -EFAULT;

    if (__put_user(hsid_read_pr(sid, i << 8), buf) < 0)
        return -EFAULT;

    *ppos = i + 1;
    return 1;
}

static int hsid_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
             unsigned long arg)
{
    uint32_t parameter;
    int t;
    sid_d* sid = PRIVATEDATA(file);

    switch (cmd) {
      case HSID_IOCTL_RESET:
        hsid_reset_stream(sid, 15);
        break;

      case HSID_IOCTL_FIFOSIZE:
        /* XXX fifosize is not the size of buffer in the hardsid protocol */
        return put_user(CW4_BUFFER_SIZE, (int*)arg);

      case HSID_IOCTL_FIFOFREE:
        /* XXX the value is not the real free value, and it's not in right units */
        t = sid->hwbuff;
        return put_user(t, (int*)arg);

      case HSID_IOCTL_SIDTYPE:
        return put_user(sid->type, (int*)arg);

      case HSID_IOCTL_MUTE:
        printk("<1>" HSID_NAME ": warning: mute not supported.\n");
        break;

      case HSID_IOCTL_NOFILTER:
        sid->nofilter = arg;
        break;

      case HSID_IOCTL_FLUSH:
        while (inb(sid->port + 0xc4) != 0x00) {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(1);
        }
        break;

      case HSID_IOCTL_DELAY:
        hsid_delay(sid, arg);
        break;

      case HSID_IOCTL_READ:
        if (get_user(parameter, (int *) arg))
            return -EFAULT;
        return put_user(hsid_read_pr(sid, parameter), (int *) arg);

      case HSID_IOCTL_DEVICES:
        return sid_numSIDs;

      case HSID_IOCTL_RELEASE:
        return hsid_release_pr(file);

      case HSID_IOCTL_ALLOCATED:
        return sid ? 2 : 0;

      default:
        printk(KERN_ERR HSID_NAME ": unknown ioctl %x\n", cmd);
        return -EINVAL;
    }
    return 0;
}

/* the SID device may be opened any number of times, but only one
 * of the users can issue writes */
static int hsid_open(struct inode *inode, struct file *file)
{
    sid_d* sid = NULL;
    
    spin_lock(&hsid_open_lock);
    if (sid_open)
        goto out_busy;
    sid_open++;
    spin_unlock(&hsid_open_lock);

    sid = sid_data[0];
    sid->hwbuff = 0;
    sid->nofilter = 0;

    PRIVATEDATA(file) = sid;
    //hsid_reset_stream(sid, 15);
    return 0;

 out_busy:
    spin_unlock(&hsid_open_lock);
    return -EBUSY;
}

static int hsid_fasync(int fd, struct file *filp, int on)
{
    return fasync_helper(fd, filp, on, &hsid_async_queue);
}

static int hsid_release_pr(struct file *file)
{
    sid_d* sid = PRIVATEDATA(file);

    hsid_reset_stream(sid, 15);
    spin_lock(&hsid_open_lock);
    sid_open--;
    spin_unlock(&hsid_open_lock);
    
    return 0;
}

static int hsid_release(struct inode *inode, struct file *file)
{
    return hsid_release_pr(file);
}

/*
 *    The various file operations we support.
 */

static struct file_operations hsid_fops = 
{
    .owner   = THIS_MODULE,
    .llseek  = hsid_llseek,
    .read    = hsid_read,
    .write   = hsid_write,
#if 1
    .compat_ioctl   = hsid_ioctl,
    /*.unlocked_ioctl   = hsid_ioctl, */
#else
    .ioctl   = hsid_ioctl,
#endif
    .open    = hsid_open,
    .release = hsid_release,
    .fasync  = hsid_fasync,
};

static struct miscdevice hsid_miscdev =
{
    .name       = HSID_NAME,
    .minor      = MISC_DYNAMIC_MINOR,
    .fops       = &hsid_fops,
};

static sid_type hsid_detect(sid_d *sid)
{
    int i, val;

    /* Reset the chip */
    hsid_reset_poke(sid, 0);

    /* Set frequency */
    sid->poke(sid, 0xf, 0xff);
    udelay(4);

    /* Set TEST bit to reset the noise generator */
    sid->poke(sid, 0x12, 0x88);
    udelay(10);

    /* Clear TEST */
    sid->poke(sid, 0x12, 0x80);
    udelay(50);
    
    /* Read the current oscillator 3 output */
    val = sid->peek(sid, 0x1b);
    for (i = 0; i < 0xffff; i++) {
        /* If the value changes, we might have a SID here */
        if (sid->peek(sid, 0x1b) != val)
            break;
    }
    /* If we looped all the way, there is no SID */
    if (i == 0xffff)
        return SID_NONE;

    hsid_reset_poke(sid, 0);

    sid->poke(sid, 0xf, 0xff);
    udelay(4);

    /* Set combined waveform which doesn't work on 6581 */
    sid->poke(sid, 0x12, 0x30);
    udelay(50);
    
    for (i = 0; i < 0xffff; i++) {
        if ((sid->peek(sid, 0x1b) & 0x80) != 0)
            break;
    }
    hsid_reset_poke(sid, 15);

    /* If the previous loop didn't finish, we have a 8580, otherwise 6581 */
    if (i == 0xffff)
        return SID_6581;
    else
        return SID_8580;
}

static int hsid_detect_chips (sid_d *sid, int chips)
{
    int i, detected = 0;
    sid_d *newsid;
    for (i = 0; i < chips; i++) {
        sid->chip = (unsigned char) i;
        sid->type = hsid_detect(sid);
        if (sid->type == SID_NONE)
            continue;

        newsid = sid_data[sid_numSIDs] = 
                 kmalloc(sizeof(sid_d), GFP_KERNEL);
        memcpy (newsid, sid, sizeof (sid_d));
        sid_numSIDs++;

        /* Only store device details on first (master) sid */
        if (detected)
            newsid->pcidev = NULL;

        printk(KERN_INFO "%s card with %s as chip %d detected @ %#x\n",
               sid_card[sid->card], sid->type == SID_6581? "6581":"8580",
               i,
               sid->port);
        detected++;
    }
    return detected;
}

static irqreturn_t hsid_interrupt(int irq, void *dev_id)
{
    unsigned char irqstate;
    
    sid_d *sid=(sid_d*)dev_id;
    irqreturn_t handled = IRQ_NONE;
    if (! (inb(sid->port+0x7) & 0x08)) {
        spin_lock_irq(&hsid_bank_lock);
        outb(0xc1, sid->port+0x3); /* irq bank */
        irqstate = inb(sid->port+0xc4) & 0x10;
        outb(irqstate, sid->port+0xc4); // Clear our pending IRQs
        outb(0x1, sid->port+0x3);  /* FIFO bank */
        spin_unlock_irq(&hsid_bank_lock);
        if (irqstate)
            wake_up(&q_wait);
        handled = IRQ_HANDLED;
    }
    return handled;
}

static int hsid_probe_cmk4(struct pci_dev *pcidev,
                           const struct pci_device_id *pciid)
{
    int cw_iobase, err, version = -1;
    sid_d *sid = &sid_setup;
    const struct firmware *fw_entry = 0;
    
    if ((err = pci_enable_device(pcidev))) {
        printk(KERN_ERR HSID_NAME ": could not enable %s\n", sid_card[SID_CARD_CWMK4]);
        return err;
    }

    cw_iobase = pci_resource_start(pcidev, 0);

    if (! request_region(cw_iobase, 256, HSID_NAME)) {
        printk(KERN_ERR HSID_NAME ": I/O ports 0x%04x-0x%04x in use\n", cw_iobase, cw_iobase+255);
        return -EBUSY;
    }

    /* Get firmware */
    err = request_firmware(&fw_entry, "cmk4.fw", &pcidev->dev);
    if (err < 0) {
        printk(KERN_ERR HSID_NAME " @ 0x%04x: firmware 'cmk4.fw' not available or load failed\n", cw_iobase);
        goto hsid_probe_cmk4_fail;
    }
    /* Initialize PCI controller */
    outb(0xf1, cw_iobase + 0x0);
    outb(0x00, cw_iobase + 0x1);
    outb(0xe3, cw_iobase + 0x2);
    outb(0x00, cw_iobase + 0x4);
    outb(0x00, cw_iobase + 0x5);
    outb(0x00, cw_iobase + 0x29);
    outb(0x00, cw_iobase + 0x2b);

    /* Reset FPGA */
    outb(0x00, cw_iobase + 0x3);
    udelay(1000);
    outb(0x41, cw_iobase + 0x3);

    {   /* Load FPGA */
        unsigned int i = 0;

        /* Handle version number */
        if (fw_entry->size > 1)
        {
            version = (fw_entry->data[0] << 8) |
                       fw_entry->data[1];
            if (version == 0xffff)
                version =  -1;
            else
                i += 2;
        }
        
        for (; i < fw_entry->size; i++)
        {
            if (fw_entry->data[i] & 1)
                outb(0x43, cw_iobase + 0x3);
            else
                outb(0x41, cw_iobase + 0x3);
            /* Wait until FPGA ready */
            if ( !(inb(cw_iobase + 0x7) & 0x8) )
            {   /* Keep trying with a 10ms timeout */
                struct timeval stv;
                do_gettimeofday(&stv);
                while ( !(inb(cw_iobase + 0x7) & 0x8) )
                {
                    struct timeval tv;
                    do_gettimeofday(&tv);
                    if ( ((tv.tv_sec  - stv.tv_sec) * 1000000 +
                          (tv.tv_usec - stv.tv_usec))  > 10000 )
                    {
                        printk(KERN_ERR HSID_NAME " @ 0x%04x: firmware load timed out\n", cw_iobase);
                        err = -EIO;
                        goto hsid_probe_cmk4_fail;
                    }
                }
            }
            outb(fw_entry->data[i], cw_iobase + 0xc0);
        }
    }

    release_firmware(fw_entry);
    fw_entry = 0;

    if (! (inb(cw_iobase + 0x7) & 0x4)) {
        printk(KERN_ERR HSID_NAME " @ 0x%04x: firmware load failed\n", cw_iobase);
        err = -EIO;
        goto hsid_probe_cmk4_fail;
    }

    if (version >= 0) {
        printk(KERN_ERR HSID_NAME " @ 0x%04x: CMK4 firmware version %d.%03d\n", cw_iobase,
               version >> 8, version & 0xff);
    }

    /* Wait until FPGA is active.  We do this by checking for an open data bus when
     * reading the floppy status register */
    if (inb(cw_iobase + 0xe8) == 0x13) {
        struct timeval stv;
        do_gettimeofday(&stv);
        /* Keep trying with a 10 ms timeout */
        while (inb(cw_iobase + 0xe8) == 0x13) {
            struct timeval tv;
            do_gettimeofday(&tv);
            if (((tv.tv_sec - stv.tv_sec) * 1000000 +
                (tv.tv_usec - stv.tv_usec)) > 10000) {
                printk(KERN_ERR HSID_NAME " @ 0x%04x: FPGA failed to start\n", cw_iobase);
                err = -EIO;
                goto hsid_probe_cmk4_fail;
            }
        }    
    }

    /* Init the rest */
    memset (sid, 0, sizeof (sid_d));
    /* For now we have a cmk3 compatible FPGA */
    sid->peek   = hsid_peek_cwmk3;
    sid->poke   = hsid_poke_cwmk3;
    sid->port   = (unsigned short) cw_iobase;
    sid->card   = SID_CARD_CWMK4;
    sid->type   = hsid_detect(sid);
    sid->pcidev = pcidev;

    if (! hsid_detect_chips (sid, 2)) {
        printk(KERN_ERR HSID_NAME " @ 0x%04x: no SID(s) detected\n", cw_iobase);
        err = -EIO; /* Correct for multiple cards? */
        goto hsid_probe_cmk4_fail;
    }

    {
        // should be card structure passed to interrupt
        int ret = request_irq(pcidev->irq, hsid_interrupt, SA_SHIRQ,
                              HSID_NAME, sid_data[0]);
        if (ret)
            return ret;
    }

   // Enable IRQs
   outb(0xc1, cw_iobase + 0x3);  // IRQ bank
   outb(0xff, cw_iobase + 0xc4); // Clear all IRQs
   outb(0x10, cw_iobase + 0xcc); // SID fifo half full
   outb(0x08, cw_iobase + 0x5);  // FPGA IRQ enable
   outb(0x00, cw_iobase + 0x2a); // PCI controller irq active level - low

   outb(1, cw_iobase + 0x3);  // FIFO bank
   outb(1, cw_iobase + 0xc8); // FIFO transfer on
   outb(0, cw_iobase + 0xcc); // FIFO erase
   return 0;

hsid_probe_cmk4_fail:
    if (fw_entry)
        release_firmware(fw_entry);
    
    release_region(cw_iobase, 256);
    return err;    
}

static int __init hsid_probe(struct pci_dev *pcidev,
                             const struct pci_device_id *pciid)
{
    if (pcidev->vendor == PCI_VENDOR_ID_INDIVIDUAL
     && pcidev->subsystem_vendor == PCI_SUBSYSTEM_VENDOR_ID_INDIVIDUAL_CWMK4)
        return hsid_probe_cmk4(pcidev, pciid);
    return 0;
}

static int hsid_notify (struct notifier_block *this, unsigned long code,
                        void *unused)
{
    if (code == SYS_DOWN || code == SYS_HALT) {
        int i;

        for (i = 0; i < sid_numSIDs; i++) {
            /* Only master sids control the card */
            if (sid_data[i]->pcidev) {
                sid_d *sid = sid_data[i];
                if (sid->card == SID_CARD_CWMK4)
                    outb(0x22, sid->port + 0x3);
            }
        }
    }
    return NOTIFY_DONE;
}

static struct notifier_block hsid_notifier =
{
    .notifier_call = hsid_notify,
};


MODULE_DEVICE_TABLE(pci, id_table);

/* @FIXME@ Need more ?? PCI devices are hotswapable */
static struct pci_driver hsid_driver =
{
    .name     = HSID_NAME,
    .id_table = id_table,
    .probe    = hsid_probe,
};

static void __exit hsid_exit (void);

static int __init hsid_init(void)
{
    int ret;

    printk(KERN_INFO "HardSID Driver v" HSID_VERSION "\n");
    sid_numSIDs = 0;
    sid_open = 0;
    pci_allocated = 0;
    delay_drift = 0;
    spin_lock_init(&hsid_open_lock);
    spin_lock_init(&hsid_bank_lock);
    spin_lock_init(&hsid_delay_update);
    init_waitqueue_head(&q_wait);

    /* Once registered always leave them that way till module is unloaded.
     * This allows support for PCI hotswappable support */
    ret = misc_register(&hsid_miscdev);
    if (ret < 0) {
        printk(KERN_ERR HSID_NAME ": misc device register failed\n");
        return -EIO;
    }

    register_reboot_notifier(&hsid_notifier);
    ret = pci_module_init(&hsid_driver);
#ifdef CONFIG_PROC_FS
    create_proc_read_entry (HSID_NAME, 0, 0, hsid_read_proc, NULL);
#endif

    pci_allocated = (ret == 0);
    if (sid_numSIDs || pci_allocated)
        return 0;

    printk(KERN_ERR HSID_NAME ": could not find any HardSID cards\n");
    /* Clean up */
    hsid_exit ();
    return ret;
}

static void __exit hsid_exit (void)
{
    int i;

    unregister_reboot_notifier(&hsid_notifier);
#ifdef CONFIG_PROC_FS
    remove_proc_entry (HSID_NAME, NULL);
#endif
    if (pci_allocated)
        pci_unregister_driver(&hsid_driver);

    for (i = 0; i < sid_numSIDs; i++) {
        /* Only master sids control the card */
        if (sid_data[i]->pcidev) {
            free_irq(sid_data[i]->pcidev->irq, sid_data[i]);
            /* Helps detect sub vender/device on next boot */
            outb(0x22, sid_data[i]->port + 0x2);
            release_region(sid_data[i]->port, 256);
            pci_disable_device(sid_data[i]->pcidev);
        }
        kfree(sid_data[i]);
    }
    misc_deregister(&hsid_miscdev);
}

/*
 *    Info exported via "/proc/hardsid".
 */

static int hsid_proc_output (char *buf)
{
    int i;
    char *p = buf;
    
    p += sprintf(p, "In use: %s\n", sid_open ? "yes" : "no");
    
    for (i = 0; i < sid_numSIDs; i += 1) {
        p += sprintf(p, "SID %d type: %s\n", i,
                sid_data[i]->type == SID_NONE ? "none" :
                sid_data[i]->type == SID_6581 ? "6581" :
                sid_data[i]->type == SID_8580 ? "8580" :
                "unknown");
    }
    if (sid_open)
        p += sprintf(p, "Hardware FIFO: %d free\n", sid_data[0]->hwbuff);

    return p - buf;
}

static int hsid_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data)
{
    int len = hsid_proc_output (page);
    
    if (len <= off+count)
        *eof = 1;
    
    *start = page + off;
    len -= off;
    
    if (len > count)
        len = count;
    if (len < 0)
        len = 0;
    
    return len;
}

module_init(hsid_init);
module_exit(hsid_exit);
MODULE_AUTHOR("Jarno Paananen, Simon White and Antti S. Lankila");
MODULE_DESCRIPTION("Driver for Catweasel Mk4 dual SID playback");
MODULE_LICENSE("GPL");
